@@ -24,7 +24,6 @@ activate = 1
 
 class OpenSSLGenerator(ConfigGenerator):
     CONFIG_NAME = 'openssl'
-    SCOPES = {'tls', 'ssl', 'openssl'}
 
     cipher_not_map = {
         'AES-256-CTR': '',
@@ -44,16 +43,6 @@ class OpenSSLGenerator(ConfigGenerator):
         'RC2-CBC': '!RC2',
         'NULL': '!eNULL:!aNULL'
     }
-    cipher_notany_multimap = {
-        # CBC is documented as SSL_DES, SSL_3DES, SSL_RC2, SSL_IDEA,
-        #                      SSL_AES128, SSL_AES256,
-        #                      SSL_CAMELLIA128, SSL_CAMELLIA256, SSL_SEED
-        '-CBC': {'DES-CBC', '3DES-CBC', 'RC2-CBC', 'IDEA-CBC',
-                 'AES-128-CBC', 'AES-256-CBC',
-                 'CAMELLIA-128-CBC', 'CAMELLIA-256-CBC', 'SEED-CBC'},
-        '-AESCCM': {'AES-128-CCM', 'AES-256-CCM'},
-        '-AESGCM': {'AES-128-GCM', 'AES-256-GCM'},
-    }
 
     key_exchange_map = {
         'RSA': 'kRSA',
@@ -87,44 +76,13 @@ class OpenSSLGenerator(ConfigGenerator):
     }
 
     ciphersuite_map = {
-        'TLS_AES_256_GCM_SHA384': {
-            'cipher': {'AES-256-GCM'},
-            'hash': {'SHA2-384'},
-            'protocol': {'TLS1.3', 'DTLS1.3'},
-        },
-        'TLS_AES_128_GCM_SHA256': {
-            'cipher': {'AES-128-GCM'},
-            'hash': {'SHA2-256'},
-            'protocol': {'TLS1.3', 'DTLS1.3'},
-        },
-        'TLS_CHACHA20_POLY1305_SHA256': {
-            'cipher': {'CHACHA20-POLY1305'},
-            'hash': {'SHA2-256'},
-            'protocol': {'TLS1.3', 'DTLS1.3'},
-        },
-        'TLS_AES_128_CCM_SHA256': {
-            'cipher': {'AES-128-CCM'},
-            'hash': {'SHA2-256'},
-            'protocol': {'TLS1.3', 'DTLS1.3'},
-        },
-        # This one is not enableable since c-p does not expose CCM8 ciphers:
-        # 'TLS_AES_128_CCM_8_SHA256': {
-        #     'cipher': {'AES-128-CCM8'},  # this is not a thing in c-p
-        #     'hash': {'SHA2-256'},
-        #     'protocol': {'TLS1.3', 'DTLS1.3'},
-        # },
-        'TLS_SHA256_SHA256': {
-            'cipher': {'NULL'},
-            'hash': {'SHA2-256'},
-            'mac': {'HMAC-SHA2-256'},
-            'protocol': {'TLS1.3', 'DTLS1.3'},
-        },
-        'TLS_SHA384_SHA384': {
-            'cipher': {'NULL'},
-            'hash': {'SHA2-384'},
-            'mac': {'HMAC-SHA2-384'},
-            'protocol': {'TLS1.3', 'DTLS1.3'},
-        },
+        'AES-256-GCM': 'TLS_AES_256_GCM_SHA384',
+        'AES-128-GCM': 'TLS_AES_128_GCM_SHA256',
+        'CHACHA20-POLY1305': 'TLS_CHACHA20_POLY1305_SHA256',
+        'AES-128-CCM': 'TLS_AES_128_CCM_SHA256',
+        'AES-128-CCM8': 'TLS_AES_128_CCM_8_SHA256',
+        'GOST28147-TC26Z-CNT': 'GOST2012-GOST8912-GOST8912',
+        'GOST28147-CPA-CNT': 'GOST2001-GOST89-GOST89'
     }
 
     @classmethod
@@ -161,9 +119,9 @@ class OpenSSLGenerator(ConfigGenerator):
                 s = cls.append(s, cls.cipher_not_map[i])
             except KeyError:
                 pass
-        for keyword, cipherset in cls.cipher_notany_multimap.items():
-            if all(c in ip['cipher'] for c in cipherset):
-                s = cls.append(s, keyword)
+        if 'AES-128-CCM' in ip['cipher']:
+            if 'AES-256-CCM' in ip['cipher']:
+                s = cls.append(s, '-AESCCM')
 
         for i in ip['mac']:
             try:
@@ -183,20 +141,17 @@ class OpenSSLGenerator(ConfigGenerator):
     def generate_ciphersuites(cls, policy):
         s = ''
         p = policy.enabled
-        # we need the output sorted by order of `cipher`
-        # finer sorting nuances are currently ignored
-        for c in p['cipher']:
-            cipher_submap = {name: spec for name, spec
-                             in cls.ciphersuite_map.items()
-                             if spec['cipher'] == {c}}
-            for ciphersuite_name, ciphersuite_spec in cipher_submap.items():
-                if all(any(val in algvalues for val in p[algclass])
-                       for algclass, algvalues in ciphersuite_spec.items()):
-                    s = cls.append(s, ciphersuite_name)
+        for i in p['cipher']:
+            try:
+                s = cls.append(s, cls.ciphersuite_map[i])
+            except KeyError:
+                pass
+
         return s
 
     @classmethod
-    def generate_config(cls, policy):
+    def generate_config(cls, unscoped_policy):
+        policy = unscoped_policy.scoped({'tls', 'ssl', 'openssl'})
         return cls.generate_ciphers(policy) + '\n'
 
     @classmethod
@@ -259,18 +214,55 @@ class OpenSSLConfigGenerator(OpenSSLGenerator):
         'RSA-PSS-RSAE-SHA2-512': 'rsa_pss_rsae_sha512',
         'EDDSA-ED25519': 'ed25519',
         'EDDSA-ED448': 'ed448',
+        # optional to not hard-require newer openssl version
+        'ECDSA-BRAINPOOLP256-SHA2-256': '?ecdsa_brainpoolP256r1_sha256',
+        'ECDSA-BRAINPOOLP384-SHA2-384': '?ecdsa_brainpoolP384r1_sha384',
+        'ECDSA-BRAINPOOLP512-SHA2-512': '?ecdsa_brainpoolP512r1_sha512',
+        # provider-only, so, optional (openssl#23050) + marked experimental
         'MLDSA44': '?mldsa44',
+        'P256-MLDSA44': '?p256_mldsa44',
+        'RSA3072-MLDSA44': '?rsa3072_mldsa44',
+        'MLDSA44-PSS2048': '?mldsa44_pss2048',
+        'MLDSA44-RSA2048': '?mldsa44_rsa2048',
+        'MLDSA44-ED25519': '?mldsa44_ed25519',
+        'MLDSA44-P256': '?mldsa44_p256',
+        'MLDSA44-BP256': '?mldsa44_bp256',
         'MLDSA65': '?mldsa65',
+        'P384-MLDSA65': '?p384_mldsa65',
+        'MLDSA65-PSS3072': '?mldsa65_pss3072',
+        'MLDSA65-RSA3072': '?mldsa65_rsa3072',
+        'MLDSA65-P256': '?mldsa65_p256',
+        'MLDSA65-BP256': '?mldsa65_bp256',
+        'MLDSA65-ED25519': '?mldsa65_ed25519',
         'MLDSA87': '?mldsa87',
+        'P521-MLDSA87': '?p521_mldsa87',
+        'MLDSA87-P384': '?mldsa87_p384',
+        'MLDSA87-BP384': '?mldsa87_bp384',
+        'MLDSA87-ED448': '?mldsa87_ed448',
+        'FALCON512': '?falcon512',
+        'P256-FALCON512': '?p256_falcon512',
+        'RSA3072-FALCON512': '?rsa3072_falcon512',
+        'FALCONPADDED512': '?falconpadded512',
+        'P256-FALCONPADDED512': '?p256_falconpadded512',
+        'RSA3072-FALCONPADDED512': '?rsa3072_falconpadded512',
+        'FALCON1024': '?falcon1024',
+        'P521-FALCON1024': '?p521_falcon1024',
+        'FALCONPADDED1024': '?falconpadded1024',
+        'P521-FALCONPADDED1024': '?p521_falconpadded1024',
+        'SPHINCSSHA2128FSIMPLE': '?sphincssha2128fsimple',
+        'P256-SPHINCSSHA2128FSIMPLE': '?p256_sphincssha2128fsimple',
+        'RSA3072-SPHINCSSHA2128FSIMPLE': '?rsa3072_sphincssha2128fsimple',
+        'SPHINCSSHA2128SSIMPLE': '?sphincssha2128ssimple',
+        'P256-SPHINCSSHA2128SSIMPLE': '?p256_sphincssha2128ssimple',
+        'RSA3072-SPHINCSSHA2128SSIMPLE': '?rsa3072_sphincssha2128ssimple',
+        'SPHINCSSHA2192FSIMPLE': '?sphincssha2192fsimple',
+        'P384-SPHINCSSHA2192FSIMPLE': '?p384_sphincssha2192fsimple',
+        'SPHINCSSHAKE128FSIMPLE': '?sphincsshake128fsimple',
+        'P256-SPHINCSSHAKE128FSIMPLE': '?p256_sphincsshake128fsimple',
+        'RSA3072-SPHINCSSHAKE128FSIMPLE': '?rsa3072_sphincsshake128fsimple',
     }
 
-    group_pq_map = {
-        'MLKEM768-X25519': '?X25519MLKEM768:?x25519_mlkem768',  # new/old name
-        'P256-MLKEM768': '?SecP256r1MLKEM768:?p256_mlkem768',  # new/old name
-        'P384-MLKEM1024': '?SecP384r1MLKEM1024:?p384_mlkem1024',  # new/old
-    }
-
-    group_classic_map = {
+    group_map = {
         'SECP224R1': 'secp224r1',
         'SECP256R1': 'secp256r1',
         'SECP384R1': 'secp384r1',
@@ -282,13 +274,29 @@ class OpenSSLConfigGenerator(OpenSSLGenerator):
         'FFDHE-4096': 'ffdhe4096',
         'FFDHE-6144': 'ffdhe6144',
         'FFDHE-8192': 'ffdhe8192',
-        'BRAINPOOL-P256R1': 'brainpoolP256r1',
-        'BRAINPOOL-P384R1': 'brainpoolP384r1',
-        'BRAINPOOL-P512R1': 'brainpoolP512r1',
+        'BRAINPOOL-P256R1': 'brainpoolP256r1:?brainpoolP256r1tls13',
+        'BRAINPOOL-P384R1': 'brainpoolP384r1:?brainpoolP384r1tls13',
+        'BRAINPOOL-P512R1': 'brainpoolP512r1:?brainpoolP512r1tls13',
+        # provider-only, so, optional (openssl#23050) + marked experimental
+        'KYBER768': '?kyber768',
+        'X25519-KYBER768': '?x25519_kyber768',
+        'P256-KYBER768': '?p256_kyber768',
+        'MLKEM512': '?mlkem512',
+        'P256-MLKEM512': '?p256_mlkem512',
+        'X25519-MLKEM512': '?x25519_mlkem512',
+        'MLKEM768': '?mlkem768',
+        'P384-MLKEM768': '?p384_mlkem768',
+        'X448-MLKEM768': '?x448_mlkem768',
+        'X25519-MLKEM768': '?X25519MLKEM768:?x25519_mlkem768',  # old/new name
+        'P256-MLKEM768': '?SecP256r1MLKEM768:?p256_mlkem768',  # old/new name
+        'MLKEM1024': '?mlkem1024',
+        'P521-MLKEM1024': '?p521_mlkem1024',
+        'P384-MLKEM1024': '?p384_mlkem1024',
     }
 
     @classmethod
-    def generate_config(cls, policy):
+    def generate_config(cls, unscoped_policy):
+        policy = unscoped_policy.scoped({'tls', 'ssl', 'openssl'})
         p = policy.enabled
         # This includes the seclevel
         s = f'CipherString = {cls.generate_ciphers(policy)}\n'
@@ -310,20 +318,8 @@ class OpenSSLConfigGenerator(OpenSSLGenerator):
         sig_algs = [cls.sign_map[i] for i in p['sign'] if i in cls.sign_map]
         s += 'SignatureAlgorithms = ' + ':'.join(sig_algs) + '\n'
 
-        # Separate groups into PQ and classic groups, generate them as follows:
-        # `*first_pq:rest_pq/*first_classic:rest_classic`.
-        # This way servers will prefer any PQ over any classic,
-        # and clients will send key_shares for top priority PQ
-        # and top priority classic groups
-        groups_pq = [cls.group_pq_map[i] for i in p['group']
-                     if i in cls.group_pq_map]
-        groups_classic = [cls.group_classic_map[i] for i in p['group']
-                          if i in cls.group_classic_map]
-        group_classes = (
-            (['*' + ':'.join(groups_pq)] if groups_pq else [])
-            + (['*' + ':'.join(groups_classic)] if groups_classic else [])
-        )
-        s += 'Groups = ' + '/'.join(group_classes) + '\n'
+        groups = [cls.group_map[i] for i in p['group'] if i in cls.group_map]
+        s += 'Groups = ' + ':'.join(groups) + '\n'
 
         if policy.enums['__ems'] == 'RELAX':
             s += 'Options = RHNoEnforceEMSinFIPS\n'
@@ -340,10 +336,10 @@ class OpenSSLConfigGenerator(OpenSSLGenerator):
 
 class OpenSSLFIPSGenerator(ConfigGenerator):
     CONFIG_NAME = 'openssl_fips'
-    SCOPES = {'tls', 'ssl', 'openssl'}
 
     @classmethod
-    def generate_config(cls, policy):
+    def generate_config(cls, unscoped_policy):
+        policy = unscoped_policy.scoped({'tls', 'ssl', 'openssl'})
         # OpenSSL EMS relaxation is special
         # in that it uses a separate FIPS module config
         # and, just in case, EMS is enforcing by default.
